"""
convert frequency to note and note to frequency.
currently, can only input frequency and note index, and not supporting note name and octave as input


usage example:
    input:
        import freq_note_converter

        val = freq_note_converter.from_freq(449)
        print(val.note)
    output:
        'A'
    input:
        print(val)
    output:
                    freq : 449
             note_index : 49
                    note : A
                  octave : 4
        offset_from_note : 0.351
        --------------------------------------------------
    input:
        freq_note_converter.from_note_index(49).print_me()
    output:
                    freq : 440.0
             note_index : 49
                    note : A
                  octave : 4
        offset_from_note : 0
        --------------------------------------------------
"""
from freq_note_converter import converter_code


def from_freq(freq):
    return _convert(freq=freq)


def from_note_index(note_index):
    return _convert(note_index=note_index)


def _convert(freq=None, note_index=None):
    if freq is not None:
        converter = converter_code.FreqNoteConverter(freq=freq)
    elif note_index is not None:
        converter = converter_code.FreqNoteConverter(note_index=note_index)
    else:
        print('no valid input')
    return converter



from unittest import TestCase
import random

import freq_note_converter


class Test(TestCase):
    def test_freq_from_freq(self):
        for _ in range(100):
            freq = random.randint(100, 1000)
            val = freq_note_converter.from_freq(freq)
            if val.freq != freq:
                self.fail(f'freq should be {freq} but module contains freq {val.freq}')

    def test_note_from_freq(self):
        frequencies = dict(A=440, B=493.9, C=523.3, D=587.3, E=659.3, F=698.5, G=784)
        for note, freq in frequencies.items():
            val = freq_note_converter.from_freq(freq)
            if val.note != note:
                self.fail(f'note should be {note} but module contains freq {val.note}')

    def test_from_note_index(self):
        frequencies = dict(A=49, B=51, C=52, D=54, E=56, F=57, G=59)
        for note, note_index in frequencies.items():
            val = freq_note_converter.from_note_index(note_index)
            if val.note != note:
                self.fail(f'note should be {note} but module contains freq {val.note}')



import math
from dict_aligned_print import print_dict


class FreqNoteConverter:
    def __init__(self, freq=None, note_index=None, note=None, octave=None):
        self.note, self.octave, self.offset_from_note, self.note_index, self.offset_from_note = '', 0, 0, 0, 0
        if freq is not None:
            self.freq = freq
            self.note, self.octave, self.offset_from_note, self.note_index = self.freq_to_note(freq)
        elif note is not None and octave is not None:
            print('not implemented')
        elif note_index is not None:
            self.note_index = note_index
            self.note, self.octave, self.freq = self.note_index_to_freq(note_index)
        else:
            print('no valid input')

    def __str__(self):
        d = dict(freq=self.freq,
                 note_index=self.note_index,
                 note=self.note,
                 octave=self.octave,
                 offset_from_note=round(self.offset_from_note, 3))
        return print_dict(d, return_instead_of_print=True)
        # output = "\n".join(["{: >16} : {}".format(k, v) for k, v in d.items()])
        # output += '\n' + '-' * 50 + '\n'
        # return output

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def note_index_to_note_octave(note_index):
        notes = ['A', 'A#', 'B', 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#']

        note = (note_index - 1) % len(notes)
        note = notes[note]

        octave = (note_index + 8) // len(notes)

        return note, octave

    @staticmethod
    def freq_to_note_index(freq):
        if not freq:  # no log value for 0
            freq += 1e-15
        # formula taken from https://en.wikipedia.org/wiki/Piano_key_frequencies
        note_index = 12 * math.log2(freq / 440) + 49
        offset_from_note = note_index
        note_index = round(note_index)
        offset_from_note -= note_index
        return note_index, offset_from_note

    def freq_to_note(self, freq):
        note_index, offset_from_note = self.freq_to_note_index(freq)
        note, octave = self.note_index_to_note_octave(note_index)

        return note, octave, offset_from_note, note_index

    def note_index_to_freq(self, note_index):
        note, octave = self.note_index_to_note_octave(note_index)
        freq = 2 ** ((note_index - 49) / 12) * 440
        return note, octave, freq

    def print_me(self):
        print(self.__str__(), end='')
